"""Utilities that don't fit anywhere else."""
import os
import tempfile
import sys
import json5 as json
import tiktoken
import openai
from .conversation import Conversation
from .config import baseConf
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'
import pygame  # noqa


def genconfig(conf_path=""):
    """Generate the config file if it doesn't exist already."""
    if not conf_path:
        conf_path = get_home_dir() + "/.owega.json"
    is_blank = True
    if (os.path.exists(conf_path)):
        is_blank = False
        with open(conf_path, "r") as f:
            if len(f.read()) < 2:
                is_blank = True
    if is_blank:
        with open(conf_path, "w") as f:
            f.write('// vim: set ft=json5:\n')
            f.write(json.dumps(baseConf.baseConf, indent=4))
        info_print(f"generated config file at {conf_path}!")
        return
    print(clrtxt('red', ' WARNING ')
        + f": YOU ALREADY HAVE A CONFIG FILE AT {conf_path}")
    print(clrtxt('red', ' WARNING ')
        + ": DO YOU REALLY WANT TO OVERWRITE IT???")
    inps = clrtxt("red", "   y/N   ") + ': '
    inp = input(inps).lower().strip()
    if inp:
        if inp[0] == 'y':
            with open(conf_path, "w") as f:
                f.write('// vim: set ft=json5:\n')
                f.write(json.dumps(baseConf.baseConf, indent=4))
            info_print(f"generated config file at {conf_path}!")
            return
    info_print("Sorry, not sorry OwO I won't let you nuke your config file!!!")


def play_opus(location: str) -> None:
    """Play an OPUS audio file."""
    pygame.mixer.init()
    sound = pygame.mixer.Sound(location)
    sound.play()
    try:
        while pygame.mixer.get_busy():
            pygame.time.delay(100)
    except KeyboardInterrupt:
        pass
    pygame.mixer.quit()


def play_tts(text: str, voice: str = 'nova') -> None:
    """Generate TTS audio from given text and play it."""
    tmpfile = tempfile.NamedTemporaryFile(
        prefix="owegatts.",
        suffix=".opus",
        delete=False
    )
    tmpfile.close()
    tts_answer = openai.audio.speech.create(
        model='tts-1',
        voice=voice,
        input=text
    )
    tts_answer.stream_to_file(tmpfile.name)
    play_opus(tmpfile.name)
    os.remove(tmpfile.name)


def estimated_tokens(ppt: str, messages: Conversation, functions):
    """Estimate the history tokens."""
    try:
        encoder = tiktoken.encoding_for_model("gpt-4")
        req = ""
        req += ppt
        req += json.dumps(messages.get_messages())
        req += json.dumps(functions)
        tokens = encoder.encode(req)
        return len(tokens)
    except Exception as e:
        if baseConf.baseConf.get("debug"):
            print(
                "An error has occured while estimating tokens:",
                file=sys.stderr
            )
            print(e, file=sys.stderr)
        return 0


def get_temp_file() -> str:
    """Get a temp file location."""
    tmp = tempfile.NamedTemporaryFile(
        prefix="owega_temp.",
        suffix=".json",
        delete=False
    )
    filename = tmp.name
    tmp.close()
    return filename


def get_home_dir() -> str:
    """Get the user home directory, cross platform."""
    return os.path.expanduser('~')


def clr(color: str) -> str:
    """Return the ANSI escape sequence for the given color."""
    esc = '\033['
    colors = {
        "red": f"{esc}91m",
        "green": f"{esc}92m",
        "yellow": f"{esc}93m",
        "blue": f"{esc}94m",
        "magenta": f"{esc}95m",
        "cyan": f"{esc}96m",
        "white": f"{esc}97m",
        "reset": f"{esc}0m",
    }
    return colors[color]


def clrtxt(color: str, text: str) -> str:
    """Print text in color between square brackets."""
    return "[" + clr(color) + text + clr("reset") + "]"


def debug_print(text: str, debug: bool = False) -> None:
    """Print text if debug is enabled."""
    if debug:
        print(' ' + clrtxt("magenta", " DEBUG ") + ": " + text)


def success_msg():
    """Return the standard success message."""
    return '  ' + clrtxt("cyan", " INFO ") + ": Owega exited successfully!"


def clearScreen():
    """Clear the terminal screen, depends on system (unix or windows-based)."""
    if os.name == 'nt':
        os.system('cls')
    else:
        print("\033[2J\033[0;0f", end="")


def do_quit(msg="", value=0, temp_file="", is_temp=False, should_del=False):
    """Quit and delete the given file if exists."""
    if (temp_file):
        if should_del:
            try:
                os.remove(temp_file)
            except Exception:
                pass
        else:
            if is_temp:
                try:
                    with open(temp_file, 'r') as f:
                        contents = json.loads(f.read())
                        if not (
                            (len(contents.get("messages", [])) > 0)
                            or (len(contents.get("souvenirs", [])) > 0)
                        ):
                            os.remove(temp_file)
                except Exception:
                    pass
    if (msg):
        print()
        print(msg)
    sys.exit(value)


def info_print(msg):
    """Print an info message."""
    print('  ' + clrtxt("cyan", " INFO ") + ": ", end='')
    print(msg)


def command_text(msg):
    """Print an command message."""
    return ' ' + clrtxt("red", "COMMAND") + ": " + msg


def print_help(commands_help={}):
    """Print the command help."""
    commands = list(commands_help.keys())
    longest = 0
    for command in commands:
        if len(command) > longest:
            longest = len(command)
    longest += 1
    print()
    info_print("Enter your question after the user prompt, "
        + "and it will be answered by OpenAI")
    info_print("other commands are:")
    for cmd, hstr in commands_help.items():
        command = '/' + cmd
        info_print(f"   {command:>{longest}}  - {hstr}")
    print()



"""Ask a question to GPT."""
from .config import baseConf
from .conversation import Conversation
from .OwegaFun import existingFunctions, connectLTS, functionlist_to_toollist
import time
import openai
import json5 as json
import re
from .utils import debug_print


def convert_invalid_json(invalid_json):
    """
    Try converting invalid json to valid json.

    Sometimes, GPT will give back invalid json.
    This function tries to make it valid.
    """
    def replace_content(match):
        content = match.group(1)
        content = (
            content
            .replace('"', '\\"')
            .replace("\n", "\\n")
        )
        return f'"{content}"'
    valid_json = re.sub(r'`([^`]+)`', replace_content, invalid_json)
    return valid_json


# ask openAI a question
# TODO: comment a lot more
def ask(
    prompt: str = "",
    messages: Conversation = Conversation(),
    model=baseConf.get("model", ""),
    temperature=baseConf.get("temperature", 0.8),
    max_tokens=baseConf.get("max_tokens", 3000),
    function_call="auto",
    temp_api_key="",
    temp_organization="",
    top_p=baseConf.get("top_p", 1.0),
    frequency_penalty=baseConf.get("frequency_penalty", 0.0),
    presence_penalty=baseConf.get("presence_penalty", 0.0),
):
    """Ask OpenAI a question."""
    if baseConf.get("debug", False):
        bc = baseConf.copy()
        bc["api_key"] = "REDACTED"
        debug_print(f"{bc}", True)
    connectLTS(messages.add_memory, messages.remove_memory, messages.edit_memory)
    old_api_key = openai.api_key
    old_organization = openai.organization
    if (prompt):
        messages.add_question(prompt)
    else:
        prompt = messages.last_question()
    if isinstance(function_call, bool):
        if function_call:
            function_call = "auto"
        else:
            function_call = "none"
    response = False
    while (not response):
        try:
            if (temp_api_key):
                openai.api_key = temp_api_key
            if (temp_organization):
                openai.organization = temp_organization
            if "vision" not in model:
                response = openai.chat.completions.create(
                    model=model,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    top_p=top_p,
                    frequency_penalty=frequency_penalty,
                    presence_penalty=presence_penalty,
                    messages=messages.get_messages(),
                    tools=functionlist_to_toollist(
                        existingFunctions.getEnabled()),
                    tool_choice=function_call,
                )
            else:
                response = openai.chat.completions.create(
                    model=model,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    top_p=top_p,
                    frequency_penalty=frequency_penalty,
                    presence_penalty=presence_penalty,
                    messages=messages.get_messages_vision(),
                )
            if (temp_api_key):
                openai.api_key = old_api_key
            if (temp_organization):
                openai.organization = old_organization
        except openai.BadRequestError as e:
            try:
                messages.shorten()
            except Exception:
                print("[Owega] Critical error... Aborting request...")
                print("[Owega] " +
                      "Please, send the following to @darkgeem on discord")
                print("[Owega] Along with a saved .json of your request.")
                print(e)
                return messages
        except openai.InternalServerError:
            print("[Owega] Service unavailable...", end="")
            time.sleep(1)
            print(" Retrying now...")
    # do something with the response
    message = response.choices[0].message
    while message.tool_calls is not None:
        try:
            for tool_call in message.tool_calls:
                tool_function = tool_call.function
                function_name = tool_function.name
                try:
                    kwargs = json.loads(tool_function.arguments)
                except json.decoder.JSONDecodeError:
                    unfixed = tool_function.arguments
                    fixed = convert_invalid_json(unfixed)
                    kwargs = json.loads(fixed)
                function_response = \
                    existingFunctions.getFunction(function_name)(**kwargs)
                messages.add_function(function_name, function_response)
            response2 = False
            while not (response2):
                try:
                    if (temp_api_key):
                        openai.api_key = temp_api_key
                    if (temp_organization):
                        openai.organization = temp_organization
                    response2 = openai.chat.completions.create(
                        model=model,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        top_p=top_p,
                        frequency_penalty=frequency_penalty,
                        presence_penalty=presence_penalty,
                        messages=messages.get_messages(),
                        tools=functionlist_to_toollist(
                            existingFunctions.getEnabled()),
                        tool_choice=function_call,
                    )
                    if (temp_api_key):
                        openai.api_key = old_api_key
                    if (temp_organization):
                        openai.organization = old_organization
                except openai.error.InvalidRequestError:
                    messages.shorten()
                except openai.error.ServiceUnavailableError:
                    print("[Owega] Service unavailable...", end="")
                    time.sleep(1)
                    print(" Retrying now...")
                message = response2.choices[0].message
        except Exception as e:
            print("Exception: " + str(e))
            print(message.tool_calls[0].function.name)
            print(message.tool_calls[0].function.arguments)
            break
    try:
        messages.add_answer(message.content.strip())
    except Exception as e:
        print("Exception: " + str(e))
    return messages



"""Handle /history."""
from ..OwegaSession import OwegaSession as ps


# shows chat history
def handle_history(
    temp_file, messages, given="", temp_is_temp=False, silent=False):
    """Handle /history."""
    if not silent:
        messages.print_history()
    return messages


item_history = {
    "fun": handle_history,
    "help": "prints the conversation history",
    "commands": ["history"],
}



"""Handle /genconf."""
from ..utils import genconfig
from ..OwegaSession import OwegaSession as ps


# generates config file
def handle_genconf(
    temp_file, messages, given="", temp_is_temp=False, silent=False):
    """Handle /genconf."""
    genconfig()
    return messages


item_genconf = {
    "fun": handle_genconf,
    "help": "generates a sample config file",
    "commands": ["genconf"],
}
